package be.iminds.ilabt.jfed.restapi;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.SimpleGeniUserWithSpeaksFor;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException;
import be.iminds.ilabt.jfed.restapi.jfedclient.JFedExperimenterApi;
import be.iminds.ilabt.jfed.restapi.jfedclient.JFedExperimenterApiImpl;
import be.iminds.ilabt.jfed.restapi.json.Experiment;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.util.jsonld.JsonLdObjectsMetaData;
import be.iminds.ilabt.util.jsonld.UriTool;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import io.dropwizard.Configuration;
import io.dropwizard.db.DataSourceFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import java.io.File;
import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Map;

public class JFedRestApiConfiguration extends Configuration {
    private static final Logger LOG = LoggerFactory.getLogger(JFedRestApiConfiguration.class);

    public static enum Access {
        REST_API, /* can use REST API, if a speaksfor matching the authenticated user is present. Can also register its own speaksfor */
        SPEAK_FOR_OTHERS, /* can use REST API as any user, if a speaksfor for that user is present. Can register any speaksfor.*/
        ADMIN /* skips all security checks */ };

    private final boolean test;

    @Valid
    @NotNull
    private final DataSourceFactory database;

//    private final SwaggerBundleConfiguration swaggerBundleConfiguration;

    private final String baseUrl;

    private final List<String> trustedUsers;
    private final List<String> trustedTlds;
    private final List<String> adminUsers;
    private final List<String> speakForOtherUsers;

    private final String sfaUserLoginPemFilename;

    private final Map<String, String> testuserMapping;

    @JsonIgnore
    private UriTool uriTool;

    @JsonCreator
    public JFedRestApiConfiguration(@JsonProperty("test")                             Boolean test,
                                    @NotNull @JsonProperty("database")                DataSourceFactory database,
                                    @NotNull @JsonProperty("baseUrl")                 String baseUrl,
                                    @JsonProperty("trustedUsers")                     List<String> trustedUsers,
                                    @JsonProperty("trustedTlds")                      List<String> trustedTlds,
                                    @JsonProperty("adminUsers")                       List<String> adminUsers,
                                    @JsonProperty("speakForOtherUsers")               List<String> speakForOtherUsers,
                                    @NotNull @JsonProperty("sfaUserLoginPemFilename") String sfaUserLoginPemFilename,
                                    @JsonProperty("testuserMapping") Map<String, String> testuserMapping
    ) {
        this.test = test == null ? false : test; //test defaults to false
        this.database = database == null ? new DataSourceFactory() : database;
//        this.swaggerBundleConfiguration = swaggerBundleConfiguration;
        this.baseUrl = baseUrl;
        this.trustedUsers = trustedUsers;
        this.trustedTlds = trustedTlds;
        this.adminUsers = adminUsers;
        this.speakForOtherUsers = speakForOtherUsers;
        this.sfaUserLoginPemFilename = sfaUserLoginPemFilename;
        this.testuserMapping = testuserMapping;
    }

    @JsonProperty
    public boolean isTest() {
        return test;
    }

    @JsonProperty
    public String getSfaUserLoginPemFilename() {
        return sfaUserLoginPemFilename;
    }


    @JsonProperty
    public String getBaseUrl() {
        return baseUrl;
    }


    @JsonProperty("database")
    public DataSourceFactory getDataSourceFactory() {
        return database;
    }

    /**
     * Users trusted to perform REST service calls.
     * @return
     */
    @JsonProperty
    public List<String> getTrustedUsers() {
        return trustedUsers;
    }

    /**
     * TLD's of which ALL users are trusted to perform service calls.
     * @return
     */
    @JsonProperty
    public List<String> getTrustedTlds() {
        return trustedTlds;
    }

    /**
     * Users trusted with admin access. (this skips almost all security checks directly)
     * @return
     */
    @JsonProperty
    public List<String> getAdminUsers() {
        return adminUsers;
    }

    /**
     * Users trusted to speak for any other user. (this indirectly skips almost all security checks)
     * @return
     */
    @JsonProperty
    public List<String> getSpeakForOtherUsers() {
        return speakForOtherUsers;
    }

    /**
     * This is a hack for debugging.
     * It allows authenticated users to be mapped to other users for internal authentication and authorization purposes.
     *
     * For example, you could map a dedicated test user to a real user account
     */
    @JsonProperty
    public Map<String, String> getTestuserMapping() {
        return testuserMapping;
    }

    private JFedExperimenterApi jFedExperimenterApi;
    @JsonIgnore
    public JFedExperimenterApi getJFedExperimenterApi() {
        if (jFedExperimenterApi == null) {
            if (sfaUserLoginPemFilename == null) {
                LOG.error("Option sfaUserLoginPemFilename is not set");
                return null;
            }
            try {
                jFedExperimenterApi = new JFedExperimenterApiImpl(new File(sfaUserLoginPemFilename), null,
                        null/*s4creds*/, null/*s4user*/);
            } catch (IOException e) {
                LOG.error("Failed to use user login PEM file \""+sfaUserLoginPemFilename+"\"", e);
                return null;
            } catch (SimpleGeniUserWithSpeaksFor.InvalidSpeaksForException e) {
                LOG.error("Invalid speaksFor credential", e);
                return null;
            } catch (InvalidLoginException e) {
                LOG.error("Login failed. (login PEM file might be wrong or expired)", e);
                return null;
            }
        }
        return jFedExperimenterApi;
    }

    private GeniUser user;
    @JsonIgnore
    public GeniUser getUser() {
        if (user == null) {
            if (sfaUserLoginPemFilename == null) {
                LOG.error("Option sfaUserLoginPemFilename is not set");
                return null;
            }

            JFedExperimenterApi jFedExperimenterApi1 = getJFedExperimenterApi();
            this.user = jFedExperimenterApi1.getUser();
        }
        return user;
    }

    private static JsonLdObjectsMetaData instance = null;
    @JsonIgnore
    public  static JsonLdObjectsMetaData getJsonLdObjectsMetaData() {
        if (instance == null) {
            instance = JsonLdObjectsMetaData.getInstance(Experiment.class.getPackage().getName());
        }
        return instance;
    }

    @JsonIgnore
    public UriTool getUriTool() {
        assert baseUrl != null;
        if (uriTool == null) {
            uriTool = getJsonLdObjectsMetaData().makeUriTool(baseUrl, new JFedRestDerivedUriGenerator());
        }
        return uriTool;
    }



    public void assureAccessAllowed(Access access, HttpServletRequest request) throws WebApplicationException {
        if (!hasAccess(access, request))
            throw new WebApplicationException("Permission denied", Response.Status.FORBIDDEN);
    }

    public boolean hasAccess(Access access, GeniUrn user) {
        if (user == null)
            return false;

        if (access == Access.REST_API) {
            String userTld = user.getTopLevelAuthority();
            LOG.debug("Checking if "+user+" tld ("+userTld+") is in trustedTlds list "+trustedTlds);
            if (trustedTlds != null && !trustedTlds.isEmpty()) {
                for (String trustedTld : trustedTlds) {
                    if (userTld.equals(trustedTld.trim()))
                        return true;
                }
            }
            LOG.debug("Checking if "+user+" is in trusted list "+trustedUsers);
            if (trustedUsers != null && !trustedUsers.isEmpty()) {
                for (String trustedUser : trustedUsers) {
                    GeniUrn trustedUserUrn = GeniUrn.parse(trustedUser.trim());
                    if (trustedUserUrn != null) {
                        if (user.equals(trustedUserUrn))
                            return true;
                    }
                }
            }
        }
        if (access == Access.ADMIN || access == Access.REST_API) { //admins also have REST_API access
            LOG.debug("Checking if "+user+" is in admin list "+adminUsers);
            if (adminUsers != null && !adminUsers.isEmpty()) {
                for (String adminUser : adminUsers) {
                    GeniUrn adminUserUrn = GeniUrn.parse(adminUser.trim());
                    if (adminUserUrn != null) {
                        if (user.equals(adminUserUrn))
                            return true;
                    }
                }
            }
        }
        if (access == Access.SPEAK_FOR_OTHERS || access == Access.REST_API) { //SPEAK_FOR_OTHERS users also have REST_API access
            LOG.debug("Checking if "+user+" is in speakForOthers list "+speakForOtherUsers);
            if (speakForOtherUsers != null && !speakForOtherUsers.isEmpty()) {
                for (String speakForOtherUser : speakForOtherUsers) {
                    GeniUrn speakForOtherUserUrn = GeniUrn.parse(speakForOtherUser.trim());
                    if (speakForOtherUserUrn != null) {
                        if (user.equals(speakForOtherUserUrn))
                            return true;
                    }
                }
            }
        }

        return false;
    }

    public static GeniUrn getAuthenticatedUser(HttpServletRequest request) {
        java.security.cert.X509Certificate[] clientCerts =
                (java.security.cert.X509Certificate[]) request.getAttribute("javax.servlet.request.X509Certificate");
        if (clientCerts != null) {
            if (clientCerts.length == 0) {
                LOG.debug("request client cert = empty");
                return null;
            } else {
                //                    for (int i = 0; i < clientCerts.length; i++)
                //                        LOG.debug("request client cert " + i + "= " + clientCerts[i].getSubjectX500Principal().getName());
                X509Certificate firstCert = clientCerts[0];
                List<GeniUrn> userUrns = KeyUtil.findUrnsInCertAltNames(
                        firstCert,
                        KeyUtil.AltNamesSource.SUBJECT_ALT_NAMES,
                        true /*users only*/);
                if (!userUrns.isEmpty()) {
                    //TODO: for security, we should verify that the issuer is trusted as authority for these users!
                    GeniUrn authenticatedUser = userUrns.get(0);
                    LOG.debug("request client cert processed -> Successfully Authenticated as "+authenticatedUser);
                    if (userUrns.size() > 1)
                        LOG.debug("request client cert note: there were additional user URNs in subject alt names. All user urns: "+userUrns);
                    return authenticatedUser;
                } else {
                    LOG.debug("request client cert contains no user URNs in the subject al names. (principal name="+firstCert.getSubjectX500Principal().getName()+")");
                    return null;
                }
            }
        } else {
            LOG.debug("request client cert = null");
            return null;
        }
    }

    public boolean hasAccess(Access access, HttpServletRequest request) {
        if (request != null) {
            String remoteAddress = request.getRemoteAddr();
            boolean isLocalhost = remoteAddress != null &&
                    (remoteAddress.equals("0:0:0:0:0:0:0:1") ||
                            remoteAddress.equals("::1") ||
                            remoteAddress.startsWith("127.0.0.1") ||
                            remoteAddress.equals("localhost"));
            LOG.debug("Got request from \""+remoteAddress+"\". isLocalhost="+isLocalhost);

            GeniUrn authenticatedUser = getAuthenticatedUser(request);
            if (authenticatedUser != null) {
                LOG.debug("request client cert processed -> Successfully Authenticated as "+authenticatedUser);
                return hasAccess(access, authenticatedUser);
            } else {
                LOG.debug("request has no authenticatedUser");
                return false;
            }
        } else {
            LOG.warn("@Context HttpServletRequest request == null. (Are you testing?)");
            return false;
        }
    }

    /**
     * Change the authenticated user, as if the required authorized user is authenticated.
     * This only occurs for users with the special SPEAK_FOR_OTHERS access.
     * (ADMINs don't use this, they bypass security more directly)
     *
     * @param request
     * @param originalAuthenticatedUser
     * @param requiredAuthorizedUser
     * @return
     */
    public GeniUrn modifyAuthenticatedUser(HttpServletRequest request, GeniUrn originalAuthenticatedUser, GeniUrn requiredAuthorizedUser) {
        if (hasAccess(Access.SPEAK_FOR_OTHERS, request)) {
            LOG.debug("modifyAuthenticatedUser causes authenticated user "+originalAuthenticatedUser+" to now appear to be authenticated as "+requiredAuthorizedUser);
            return requiredAuthorizedUser;
        }
        return originalAuthenticatedUser;
    }

//    @JsonProperty("swagger")
//    public SwaggerBundleConfiguration getSwaggerBundleConfiguration() {
//        return swaggerBundleConfiguration;
//    }
}
